#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <linux/ioctl.h>
#include <asm/uaccess.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/module.h>
#include <mach/mt_gpio.h>
#include <mach/mt_gpio_base.h>
#include <mach/mt_gpio.h>
#include <mach/eint.h>
#include <mach/mt_gpt.h>
#include <cust_gpio_usage.h>

#include "common.h"
#include "hdvd_tray.h"

#if FLAG_IS_MODEL
#define SONY_HDVD_TRAYOUT_EINT 161
#define SONY_HDVD_TRAYIN_EINT 77
#else
#define SONY_HDVD_TRAYOUT_EINT 105
#define SONY_HDVD_TRAYIN_EINT 101
#endif
#define OPENING_BREAK 20
#define CLOSING_BREAK 30

static E_TRAY_STATE	e_tray_state = E_NOP;
static E_TRAY_POSITION	e_tray_position = E_UNKNOWNPOS;
static int m_target_pos = 0;
static int m_gpt_flag = 0;
static int m_timer_flag = 0;
static unsigned long timecount = 0;
static struct timer_list out_break_timer;
static struct timer_list close_break_timer;
static  int m_act_end;

// private data for each DVD tray driver LSI
struct priv {
	struct miscdevice	misc;			// MISC device
};

static struct class	*myclass;
static struct cdev	char_dev;
static dev_t		dev;

static int	jiffies_count_over(unsigned long start, unsigned long duration)
{
	unsigned long	target = start + duration;

	if (target >= start)
		return (target <= jiffies);
	return (jiffies < start && jiffies >= target);
}

static void	my_wait_jiffies_count(long timeout)
{
	const unsigned long	start_jiffy = jiffies;

	while (!jiffies_count_over(start_jiffy, timeout)) {
		my_jiffy_sleep(1);
	}
}


static E_TRAY_POSITION	hdvd_tray_getpos(void)
{
#define	BIT_TRAY_IN	1
#define	BIT_TRAY_OUT	2

	int	status = 0;

	if (mt_get_gpio_in(SONY_HDVD_TRAYOUT_PIN))
		status |= BIT_TRAY_OUT;

	if (mt_get_gpio_in(SONY_HDVD_TRAYIN_PIN))
		status |= BIT_TRAY_IN;

	//DBGTRACE(" sensor status %d", status);

	switch (status) {
	case (BIT_TRAY_IN | BIT_TRAY_OUT):
		return E_OPENED;

	case 0:
		return E_CLOSED;

	case BIT_TRAY_IN:
		return E_HALFOPENED;

	case BIT_TRAY_OUT:
	default:
		return E_UNKNOWNPOS;
	}

	return E_UNKNOWNPOS;	 // sequence must not reach here
}

static void	hdvd_tray_setmotormode(E_TRAY_STATE e_state)
{
	switch (e_state) {
	case E_OPEN:
		mt_set_gpio_out(SONY_HDVD_FWD_PIN, GPIO_OUT_ZERO);
		mt_set_gpio_out(SONY_HDVD_REV_PIN, GPIO_OUT_ONE);
		break;
	case E_CLOSE:
		mt_set_gpio_out(SONY_HDVD_FWD_PIN, GPIO_OUT_ONE);
		mt_set_gpio_out(SONY_HDVD_REV_PIN, GPIO_OUT_ZERO);
		break;
	case E_BRAKE:
		mt_set_gpio_out(SONY_HDVD_FWD_PIN, GPIO_OUT_ONE);
		mt_set_gpio_out(SONY_HDVD_REV_PIN, GPIO_OUT_ONE);
		break;
	case E_NOP:
	default:	 // switch-case must not match default
		mt_set_gpio_out(SONY_HDVD_FWD_PIN, GPIO_OUT_ZERO);
		mt_set_gpio_out(SONY_HDVD_REV_PIN, GPIO_OUT_ZERO);
		e_state = E_NOP;
		break;
	}
	e_tray_state = e_state;
}
static void hdvd_tray_open_pwm( unsigned long id )
{
	static unsigned long count;
	
	timecount++;
	
	// first in initial count
	if( timecount == 1 )
	{
		count = 0;
		hdvd_tray_setmotormode(m_target_pos);
	}
	
	count++;
	
	// 50msec pass
	if( count == 5 )
	{
		count = 0;
		hdvd_tray_setmotormode(m_target_pos);
	}
	
	// 40msec pass
	else if( count == 4 )
	{
		hdvd_tray_setmotormode(E_NOP);
	}
	
}

static void hdvd_tray_close_first_pwm( unsigned long time )
{
	
	switch(time)
	{
		case 1: // 10msec pass
		case 2: // 20msec pass
		case 4: // 30msec pass
		case 5:
		case 7:
		case 8:
		
		hdvd_tray_setmotormode(E_NOP);
		break;
		
		default:
		hdvd_tray_setmotormode(m_target_pos);
		
		break;
	}
}

static void hdvd_tray_close_second_pwm( unsigned long time )
{
	
	switch(time)
	{
		case 10:
		case 12:
		case 14:
		case 16:
		
		hdvd_tray_setmotormode(E_NOP);
		break;
		
		default:
		
		hdvd_tray_setmotormode(m_target_pos);
		
		break;
	}
}


static void hdvd_tray_close_third_pwm( unsigned long time )
{
	
	switch(time)
	{
		case 19:
		case 22:
		case 25:
		
		hdvd_tray_setmotormode(E_NOP);
		break;
		
		default:
		
		hdvd_tray_setmotormode(m_target_pos);
		
		break;
	}
}


static void hdvd_tray_close_four_pwm( unsigned long time )
{
	static unsigned long count;
	
	count++;
	
	// first in initial count
	if( time == 1 )
	{
		count = 1;
		hdvd_tray_setmotormode(m_target_pos);
	}
	
	// 50msec pass
	if( count == 5 )
	{
		count = 0;
		hdvd_tray_setmotormode(m_target_pos);
	}
	
	// 40msec pass
	if( count == 4 )
	{
		hdvd_tray_setmotormode(E_NOP);
	}
	
}

static void hdvd_tray_close_pwm( unsigned long id )
{
	timecount++;
	
	if( timecount < 10 ) // first pwm 0`90msec pass
	{
		hdvd_tray_close_first_pwm( timecount );
	}
	
	else if( timecount < 18 ) // second pwm 100`170msec pass
	{
		hdvd_tray_close_second_pwm( timecount );
	}
	
	else if( timecount < 27 ) // second pwm 180`260msec pass
	{
		hdvd_tray_close_third_pwm( timecount );
	}
	
	else  // third pwm 270msec`
	{
		hdvd_tray_close_four_pwm( timecount -26 );
	}
}
static void hdvd_tray_pwm( unsigned long id )
{
	if( m_target_pos == E_OPENED )
	{
		hdvd_tray_open_pwm(id);
	}
	else
	{
		hdvd_tray_close_pwm(id);
	}
}
static void break_timer_handler( unsigned long data )
{
	m_act_end = 1;
	hdvd_tray_setmotormode(E_NOP);
}

static void hdvd_tray_in_handler( void )
{
	if( ( m_act_end != 0 ) || ( m_target_pos != E_CLOSED ) )
	{
		return;
	}
	hdvd_tray_setmotormode(E_BRAKE);
	
	stop_gpt( GPT2 ) ;
	m_gpt_flag = 0;
	m_timer_flag = 1;
	del_timer( &close_break_timer );
	add_timer( &close_break_timer );

}

static void hdvd_tray_out_handler( void )
{
	
	if( ( m_act_end != 0 ) || ( m_target_pos != E_OPENED ) )
	{
		return;
	}
	
	hdvd_tray_setmotormode(E_BRAKE);
	
	stop_gpt( GPT2 ) ;
	m_gpt_flag = 0;
	m_timer_flag = 1;
	del_timer( &out_break_timer );
    add_timer( &out_break_timer );
	
}
static void	hdvd_tray_cancel()
{
	hdvd_tray_setmotormode(E_NOP);
	if( !m_act_end )
	{
		m_act_end = 1;
		
		if( m_gpt_flag )
		{
			stop_gpt( GPT2 ) ;
		}
		
		if( m_target_pos == E_OPENED )
		{
			del_timer(&out_break_timer);
		}
		else
		{
			del_timer(&close_break_timer);
		}
	}
}
static int	hdvd_tray_doaction(E_TRAY_STATE e_state)
{
	E_TRAY_POSITION	e_target_pos;
	int		pwm_count = 0;
	unsigned long	start_jiffy;
	unsigned long	timer;
	int result = 0;

	switch (e_state) {
	case E_OPEN:
		e_target_pos = E_OPENED;
		timer = TRAY_OPEN_TIMEOUT;
		break;
	case E_CLOSE:
		e_target_pos = E_CLOSED;
		timer = TRAY_CLOSE_TIMEOUT;
		break;
	case E_END:
		e_target_pos = E_CLOSED;
		timer = 0;
		break;
	default:
		return;
	}

	start_jiffy = jiffies;
	m_act_end = 0;
	timecount = 0;
	m_target_pos = e_target_pos;
	m_gpt_flag = 0;
	m_timer_flag = 0;
	
	
	while (!m_act_end)
	{
		if ( ( jiffies_count_over(start_jiffy, timer) ) && ( timer != 0 ) )
		{
			result = ETIME;
			hdvd_tray_setmotormode(E_NOP);
			break;
		}
		if( ( m_gpt_flag ) || ( m_timer_flag ) )
		{
			msleep(10);
			continue;
		}
		
		if(e_target_pos == E_OPENED)
		{
			m_gpt_flag = 1;
			hdvd_tray_setmotormode(m_target_pos);
			start_gpt( GPT2 ) ;
			mt_eint_unmask( SONY_HDVD_TRAYOUT_EINT );
			
			continue;
		}
		
		if(e_target_pos == E_CLOSED)
		{
			m_gpt_flag = 1;
			hdvd_tray_setmotormode(m_target_pos);
			start_gpt( GPT2 ) ;
			mt_eint_unmask( SONY_HDVD_TRAYIN_EINT );
			continue;
		}
	}
	if( m_gpt_flag )
	{
		stop_gpt( GPT2 ) ;
	}
	
	mt_eint_mask(SONY_HDVD_TRAYIN_EINT);
	mt_eint_mask(SONY_HDVD_TRAYOUT_EINT);
	del_timer(&out_break_timer);
	del_timer(&close_break_timer);
	
	m_timer_flag = 0;
	timecount = 0;
	m_gpt_flag = 0;
	return result;
}

/*
 * open DVD tray  driver
 *
 *   IN : p_inode	= inode information address
 *        p_file	= driver file information
 *   OUT: always 0(means Ok)
 */
static int	hdvd_tray_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data,
					    struct priv, misc);

	return 0;
}

/*
 * DVD tray ioctl()
 *
 *  IN : file	= driver file information
 *	 cmd    = ioctl function number
 *	 arg    = ioctl have only 1 argument
 * OUT : error code
 */
static long	hdvd_tray_ioctl(struct file *file, unsigned int cmd,
				 unsigned long arg)
{
	int		err = 0;
	int		retval = 0;
	long		l;
	E_TRAY_POSITION	e_pos;

	if (!file)
		return -EINVAL;

	/* Check type and command number */
	if (_IOC_TYPE(cmd) != HDVD_TRAY_IOCTL_MAGIC)
		return -ENOTTY;

	/* Check access direction once here; don't repeat below.
	 * IOC_DIR is from the user perspective, while access_ok is
	 * from the kernel perspective; so they look reversed.
	 */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE,
				(void __user *)arg, _IOC_SIZE(cmd));
#if 0
	if (err == 0 && _IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ,
				(void __user *)arg, _IOC_SIZE(cmd));
#endif
	if (err)
		return -EFAULT;

	/* guard against device removal before, or while,
	 * we issue this ioctl.
	 */

	switch (_IOC_NR(cmd)) {
	// read request
	case _IOC_NR(HDVD_TRAY_IOC_GETPOS):
	        l = (long)hdvd_tray_getpos();
		retval = copy_to_user((void *)arg, &l, sizeof(long));
		break;

	// write requests
	case _IOC_NR(HDVD_TRAY_IOC_ACT):
		if (retval)
			break;
		retval = hdvd_tray_doaction((E_TRAY_POSITION)arg);
		break;
	case _IOC_NR(HDVD_TRAY_IOC_CANSEL):

		hdvd_tray_cancel();
		retval = 0;
		break;
	default:
		// undefined ioctl()
		retval = -EINVAL;	// sequence must not reach here
		break;
	}

	return retval;
}
/*
 * write DVD tray PWM data (test interface)
 *
 *  IN : file	= driver file information
 *	 user_bufer	= incoming data
 *	 count		= incoming data length
 *	 ppos		= write positon (ignored)
 * OUT : write count (when success)
 *	 error code
 */
static ssize_t	hdvd_tray_write(struct file *file, const char *user_buffer,
                           size_t count, loff_t *ppos)
{
	char		buffer[128];
	int		result;

	if (!file || !user_buffer || !ppos)
		return -EINVAL;

	// copy input text data
	if (count > sizeof(buffer))
		count = sizeof(buffer);
	result = copy_from_user(buffer, user_buffer, count);
	if (result)
		return result;

	// parse input text data
	switch (buffer[0]) {
	case 'O':
		// open tray with tray status check
		hdvd_tray_doaction(E_OPEN);
		break;
	case 'o':
		// open tray without motor stopping
		DBGTRACE("tray motor on to open");
		hdvd_tray_setmotormode(E_OPEN);
		break;
	case 'C':
		// close tray with tray status check
		hdvd_tray_doaction(E_CLOSE);
		break;
	case 'c':
		// close tray without motor stopping
		DBGTRACE("tray motor on to close");
		hdvd_tray_setmotormode(E_CLOSE);
		break;
	case 'S':
		// get tray status
		hdvd_tray_getpos();
		break;
	case 's':
		// stop tray motor
		DBGTRACE("tray motor off");
		hdvd_tray_setmotormode(E_NOP);
		break;
	default:
		// unknown command
		return -EINVAL;
	}

	return count;
}

static const struct file_operations	hdvd_tray_fops = {
	.owner		= THIS_MODULE,
	.open		= hdvd_tray_open,
	.unlocked_ioctl	= hdvd_tray_ioctl,
	.write		= hdvd_tray_write,
};

// initalize routine invoked when device attach (insmod)
static int __init	hdvd_tray_init(void) 
{ 
        int ret;

        dev = MKDEV( MAJOR_NUM, 0 );
        cdev_init(&char_dev, &hdvd_tray_fops );
        char_dev.owner = THIS_MODULE;
        ret = cdev_add(&char_dev, dev, 1);
        if (ret < 0)
                return ret;
        myclass = class_create( THIS_MODULE, DRV_NAME );
        device_create( myclass, NULL, dev, NULL, DRV_NAME );
	
		request_gpt(GPT2, GPT_REPEAT, GPT_CLK_SRC_SYS, GPT_CLK_DIV_13, 10000, hdvd_tray_close_pwm, 0) ;		// 13000000 Hz (GPT_CLK_SRC_SYS) / 13 (GPT_CLK_DIV_13)
	
		mt_eint_registration( SONY_HDVD_TRAYIN_EINT , EINTF_TRIGGER_LOW, hdvd_tray_in_handler, 0);
		mt_eint_mask(SONY_HDVD_TRAYIN_EINT);
		mt_eint_set_hw_debounce( SONY_HDVD_TRAYIN_EINT, 16);
		
		mt_eint_registration( SONY_HDVD_TRAYOUT_EINT , EINTF_TRIGGER_HIGH, hdvd_tray_out_handler, 0);
		mt_eint_mask(SONY_HDVD_TRAYOUT_EINT);
		mt_eint_set_hw_debounce( SONY_HDVD_TRAYOUT_EINT, 16);
		
		init_timer( &close_break_timer );
	
		close_break_timer.expires  = jiffies + CLOSING_BREAK;
		close_break_timer.data     = ( unsigned long )jiffies;
		close_break_timer.function = break_timer_handler;
	
		init_timer( &out_break_timer );
	
		out_break_timer.expires  = jiffies + OPENING_BREAK;
		out_break_timer.data     = ( unsigned long )jiffies;
		out_break_timer.function = break_timer_handler;
	

        return 0;
} 

// finalize routine invoked when device detouch (rmmod)
static void __exit	hdvd_tray_exit(void) 
{ 
	device_destroy(myclass, dev);
	class_destroy(myclass);
	if( m_gpt_flag )
	{
		stop_gpt( GPT2 ) ;
	}
	
	free_gpt( GPT2 ) ;
	del_timer(&out_break_timer);
	del_timer(&close_break_timer);
	
	mt_eint_mask(SONY_HDVD_TRAYIN_EINT);
	mt_eint_registration( SONY_HDVD_TRAYIN_EINT , EINTF_TRIGGER_LOW, NULL, 0);
	mt_eint_mask(SONY_HDVD_TRAYIN_EINT);
	
	mt_eint_mask(SONY_HDVD_TRAYOUT_EINT);
	mt_eint_registration( SONY_HDVD_TRAYOUT_EINT , EINTF_TRIGGER_HIGH, NULL, 0);
	mt_eint_mask(SONY_HDVD_TRAYOUT_EINT);

	hdvd_tray_doaction(E_NOP);
} 

module_init(hdvd_tray_init); 
module_exit(hdvd_tray_exit); 

MODULE_DESCRIPTION("model-H DVD tray Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_VERSION("v0.01");
